import functools
import typing

import iservice

P = typing.ParamSpec("P")
R = typing.TypeVar("R")
S1 = typing.TypeVar("S1", bound=iservice.Service)
S2 = typing.TypeVar("S2", bound=iservice.Service)
S3 = typing.TypeVar("S3", bound=iservice.Service)
S4 = typing.TypeVar("S4", bound=iservice.Service)


@typing.overload
def inject(
    func: typing.Callable[typing.Concatenate[type[S1], P], R],
    service1: type[S1],
    /,
) -> typing.Callable[P, R]:
    ...


@typing.overload
def inject(
    func: typing.Callable[typing.Concatenate[type[S1], type[S2], P], R],
    service1: type[S1],
    service2: type[S2],
    /,
) -> typing.Callable[P, R]:
    ...


@typing.overload
def inject(
    func: typing.Callable[typing.Concatenate[type[S1], type[S2], type[S3], P], R],
    service1: type[S1],
    service2: type[S2],
    service3: type[S3],
    /,
) -> typing.Callable[P, R]:
    ...


@typing.overload
def inject(
    func: typing.Callable[typing.Concatenate[type[S1], type[S2], type[S3], type[S4], P], R],
    service1: type[S1],
    service2: type[S2],
    service3: type[S3],
    service4: type[S4],
    /,
) -> typing.Callable[P, R]:
    ...


def inject(
    func: typing.Callable[..., R],
    /,
    *services: type[iservice.Service],
) -> typing.Callable[..., R]:
    """Read implementation accept any number of services."""
    return functools.partial(func, *services)



import abc
import typing

import pytest

import iservice


class ServiceA(iservice.Service):
    @abc.abstractmethod
    @staticmethod
    def method_a() -> str:
        """
        This method do something.
        """


class ServiceA1(ServiceA):
    @staticmethod
    def method_a() -> str:
        return "A1"


class ServiceA2(ServiceA):
    @staticmethod
    def method_a() -> str:
        return "A2"


def api_func_a(service: type[ServiceA]) -> str:
    return service.method_a()


@pytest.fixture
def api_func_a1() -> typing.Callable:
    return iservice.inject(api_func_a, ServiceA1)


@pytest.fixture
def api_func_a2() -> typing.Callable:
    return iservice.inject(api_func_a, ServiceA2)



"""
This module contains the Service Protocol, a superclass for all service subclasses.
"""
import typing


@typing.runtime_checkable
class Service(typing.Protocol):
    """
    When subclassing Service, subclass must also inherit from typing.Protocol to extend the
    protocol. This class is used for runtime check and dependency injection.
    """
